// Copyright 2003 Regents of the University of California

// Astropulse is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2, or (at your option) any later
// version.

// Astropulse is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.

// You should have received a copy of the GNU General Public License along
// with Astropulse; see the file COPYING.  If not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

/* ap_client_main.C */
/* Main program for AstroPulse client */
#include "config.h"

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef _WIN32 
#include <windows.h>
#define BOINC_APP_GRAPHICS 1
#endif

#include <cassert>
#include <cstdio>
#include <cstdlib>

#include "boinc_api.h"
#include "util.h"

#include "astropulse.h"
#include "diagnostics.h"


#ifdef BOINC_APP_GRAPHICS
#include "ap_graphics.h"
#include "graphics_api.h"
// #include "sah_gfx.h"
#endif

#include <time.h>

#ifdef _WIN32
#include <windows.h>
#include "stackwalker_win.h"

extern int main(int argc, char** argv);



int WINAPI WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR Args, int WinMode)
{
  LPSTR command_line;
  char* argv[100];
  int argc;

  //InitAllocCheck();

  command_line = GetCommandLine();
  argc = parse_command_line(command_line,argv);

  int ret = main(argc, argv);

  //DeInitAllocCheck();

  return ret;
}
#endif

/* Default thresholds */
static float thresh_def[] = {
  (float)21.50,
  (float)24.748650,
  (float)30.015163,
  (float)38.773587,
  (float)53.662645,
  (float)79.565603,
  (float)125.71365,
  (float)209.85380,
  (float)366.44671,
  (float)662.93874,
  (float)1232.2705,
  0, 0, 0, 0
};

#ifdef HAVE_GL_LIB
#ifdef UNIX

#include <GL/glut.h>

void displayCB(void)		/* function called whenever redisplay needed */
{
  glClear(GL_COLOR_BUFFER_BIT);		/* clear the display */
  glColor3f(1.0, 1.0, 1.0);		/* set current color to white */
  glBegin(GL_POLYGON);			/* draw filled triangle */
  glVertex2i(200,125);			/* specify each vertex of triangle */
  glVertex2i(100,375);
  glVertex2i(300,375);
  glEnd();				/* OpenGL draws the filled triangle */
  glFlush();				/* Complete any pending operations */
}

void keyCB(unsigned char key, int x, int y)	/* called on key press */
{
  if( key == 'q' ) exit(0);
}

#endif
#endif

void worker() {
  Astropulse::client.init();
  Astropulse::client.science.mainloop();
  Astropulse::client.finish();
}

/* Main Program */
int main(int argc, char *argv[]) {

  //char* foo = (char*) (void*) 0xbadf00d; 
  //*foo = 'x';

  //fprintf(stderr, "START TIME: %ld\n", time(NULL));

  int i, retval;
  bool standalone = false;
    
  for (i=1; i<argc; i++) { 
	if (!strcmp(argv[i], "-standalone")) {
      standalone = true;
      fprintf(stderr, "## debug standalone mode\n");
      continue;
    }
    if (!strcmp(argv[i], "-reset")) {
      fprintf(stderr, "## resetting ap_state.dat and pulse.out\n");
      unlink("ap_state.dat");
      unlink("pulse.out");
      continue;
    }
    if (standalone && !strcmp(argv[i], "-thresh0") && ++i < argc) {
      fprintf(stderr, "## thresh[0] = %f\n",
	      (Astropulse::client.state.thresh[0] =(float)atoi(argv[i])) );
      continue;
    }
  }

  // Initialize Diagnostics
  //
  unsigned long dwDiagnosticsFlags = 0;

  dwDiagnosticsFlags = 
      BOINC_DIAG_DUMPCALLSTACKENABLED | 
      BOINC_DIAG_HEAPCHECKENABLED |
      BOINC_DIAG_TRACETOSTDERR |
      BOINC_DIAG_REDIRECTSTDERR;

  retval = boinc_init_diagnostics(dwDiagnosticsFlags);
  if (retval) {
    fprintf(stderr, "boinc_init_diagnostics failed: %d\n", retval);
    exit(retval);
  }

#ifndef BOINC_APP_GRAPHICS
  retval = boinc_init();
  if (retval) {
    fprintf(stderr, "boinc_init failed: %d (try -standalone for testing)\n", retval);
    exit(retval);
  }
#else
  //app_graphics_init();
  ap_graphics.data_struct_init();
  retval = boinc_init_graphics(worker);
  if (retval) {
    fprintf(stderr, "boinc_init_graphics failed: %d\n", retval);
    exit(retval);
  }
#endif

  Astropulse::client.init();
  Astropulse::client.science.mainloop();
  Astropulse::client.finish();

  //fprintf(stderr, "END TIME: %ld\n", time(NULL));

  boinc_finish(0);
  return 0;
}

namespace Astropulse {
  Client client;

  inline int calculate_fold_level(int dm_chunk) {
    /* Figure out sizes */
    int level = 0;
    for (;; ++level) {
      int temp = dm_chunk >> level;
      if (temp == 0) {
	fprintf(stderr, "Error calculating fold_level.\n");
	exit(PARSE_ERR);
      }
      if (temp == 1) {
	return level;
      }
    }
  }

  state_t::state_t()
  {
    // Default parameters
    datasize  = 8*1024*1024;
    dm_low    = 1;                                      // ~ 0 pc/cm^3
    dm_hi     = 15000;                                  // ~ 100 pc/cm^3
    dm_chunk  = 32;                           // DM resolution for folding
    fft_len   = 32768;                 // Needs to be at least twice DM_HI
    max_coadd = 10;
    for (int l = 0; l < max_coadd; ++l) {
      thresh[l] = thresh_def[l];

      // NOTE: none of these initial values except peak_power
      // should ever be used
      best[l].peak_power = -INFINITY;
      best[l].period     = -1.0;
    }

    /* Starting place within computation */
    data_chunk_now = 0;
    dm_chunk_now   = dm_low;
    dm_now         = 0;
    dm_sign        = 1;
    fold_buf_loc   = 0;
    frac_done      = 0.0;
    result_count   = 0;
  }

  inline void Client::do_checkpoint() {
    outfile.flush();
    foldfile.write();
    statefile.write();
  }

  // This function is inlined because it's called very often.
  inline void Client::checkpoint() {
    if (boinc_time_to_checkpoint()) {
      do_checkpoint();
      boinc_checkpoint_completed();
    }
  }

  void Client::init()
  {
    // Open wu (data) file, output file
    statefile.read();
    wufile.init();
    outfile.init();
    if (!statefile.resumed) {
      // no state file exists; parse data header
      wufile.parse_header();
      state.dm_chunk_now = state.dm_low;
    }

    // Figure out sizes
    state.fold_level = calculate_fold_level(state.dm_chunk);
    state.fold_buf_size = (state.datasize*4)>>state.fold_level;

    // Init vars, fft plans, etc.
    science.init();
    power.resize(state.fft_len/2);
    fold_buf.resize(state.fold_buf_size);

    // Read/write files
    if (foldfile.read()) {
      // Couldn't read fold file => reset to beginning of dm chunk
      state.fold_buf_loc = 0;
      state.dm_now       = 0;
      state.dm_sign      = 1;
    }

    wufile.read_raw_data();
    wufile.finish();

    statefile.write();
  }

  inline void state_t::compute_fraction_done(double l)
  {
    frac_done = l/(double)max_coadd;
    frac_done += (1-dm_sign);
    frac_done /= (2+1);
    frac_done += dm_now;
    frac_done /= dm_chunk;
    frac_done += data_chunk_now/((fft_len/2)/4);
    frac_done /= (datasize/((fft_len/2)/4))-1;
    frac_done += (dm_chunk_now-dm_low)/dm_chunk;
    frac_done /= 1+((dm_hi - dm_low)/dm_chunk);
  }

  void time_series_b_t::build(float array[], int index, int range)
  {
    float peak = array[index];
    data.clear();
    if (range >= MAXLENGTH) {
      data.resize(MAXLENGTH);
      index -= MAXLENGTH/2;
      if (index < 0) {
	index = 0;
      } else if (index + MAXLENGTH > range) {
	index = range - MAXLENGTH;
      }
      for (int j = 0; j < MAXLENGTH; ++index, ++j) {
	// scale to peak, and convert to unsigned byte
	data[j] = (unsigned char) (int) (256 * array[index]/peak);
      }
      length = MAXLENGTH;
    } else {
      data.resize(range);
      for (index = 0; index < range; ++index) {
	data[index] = (unsigned char) (int) (256 * array[index]/peak);
      }
      length = range;
    }
  }

  void time_series_f_t::build(float array[], int index, int range)
  {
    data.clear();
    data.resize(MAXLENGTH);
    if (range >= MAXLENGTH) {
      index -= MAXLENGTH/2;
      if (index < 0) {
	index = 0;
      } else if (index + MAXLENGTH > range) {
	index = range - MAXLENGTH;
      }
      data.assign(array+index, array+index+MAXLENGTH);
      length = data.size();
    } else {
      data.assign(array, array+range);
      length = data.size();
    }
  }

  void Science::mainloop()
  {
    state_t& state = client.state;
    float*   power = &(client.power[0]);

    /* Main DM loop - by chunk */
    for (;state.dm_chunk_now < state.dm_hi;
	 state.dm_chunk_now += state.dm_chunk)
      {
	client.checkpoint();
	build_chirp_table(state.dm_chunk_now);
	/* Loop over all raw data (overlap) */
	for (; state.data_chunk_now<(state.datasize-state.fft_len/2/4); state.data_chunk_now+=(state.fft_len/2)/4) {
	  client.checkpoint();
	  convert_bits_to_float();
	  client.checkpoint();
	  compute_forward_fft();
	  /* Inner DM loop */
#ifdef BOINC_APP_GRAPHICS
	  double max = 0.0;
	  GRAPH_BUFFER* abp;
	  abp = ap_graphics.get_generate_buffer(true);
	  if (abp) abp->rarray.init_data(state.fft_len/2, state.dm_chunk); // const
#endif
	  for (; state.dm_now<state.dm_chunk; state.dm_now++) {
	    /* do both dm signs */
	    for (; state.dm_sign>=-1; state.dm_sign-=2) {
	      client.checkpoint();
	      dechirp(state.dm_now, power, state.dm_sign);
#ifdef BOINC_APP_GRAPHICS
	      if (abp && state.dm_sign > 0) { // add positive dispersions to graph
		abp->rarray.add_source_row(power);
	      }
#endif
	      /* loop over co-add levels */
	      for (int l=0; l<state.max_coadd; l++) {
		state.compute_fraction_done(l);
		boinc_fraction_done(state.frac_done);

		int dechirped_range_data_length = state.fft_len >> (l+1);
		for (int m=0; m<dechirped_range_data_length; m++) {  // 
#if BOINC_APP_GRAPHICS
		  if (power[m] > max) {
		    ap_graphics.time_series.build(power, m, dechirped_range_data_length);
		    max = power[m];
		  }
#endif
		  if (power[m] > state.thresh[l]) {
		    ap_signal s;
		    time_series_b_t time_series;
		    time_series.build(power, m, dechirped_range_data_length);
		    state.pulse_jd = state.jd0 + (state.data_chunk_now + state.dm_now)/(state.datasize)*(state.jd1-state.jd0);

		    s.dm= state.dm_sign*(state.dm_chunk_now+state.dm_now); //dm
	            s.scale=l;                   // scale
		    s.peak_power=power[m];       // peak_power
		    s.fft_num=state.data_chunk_now*4;
                    s.peak_bin=state.data_chunk_now * 4 + (m << l); // index
		    s.time=state.pulse_jd;       // time
		    s.period=0.0;                // period
		    s.time_series_len=time_series.length;
		    s.time_series=time_series.data;
		    s.time_series.encoding=_base64;
		    client.outfile.output_result(s);
		    if (boinc_is_standalone()) {
		      client.do_checkpoint();
		    }
		  }

		  if (power[m] > state.best[l].peak_power) {
		    state.pulse_jd = state.jd0 + (state.data_chunk_now + state.dm_now/state.dm_chunk)/(state.dm_hi)*(state.jd1-state.jd0);
		    state.best[l].fft_num=state.data_chunk_now*4;
		    state.best[l].peak_bin = state.data_chunk_now * 4 + (m << l);
		    state.best[l].peak_power = power[m];
		    state.best[l].scale = l;
		    state.best[l].dm = state.dm_sign*(state.dm_chunk_now+state.dm_now);
		    state.best[l].period = 0.0;
		    state.best[l].time = state.pulse_jd;
		  }
		} // bin number (m)
		if ((l==state.fold_level) && (state.dm_now==0) && (state.dm_sign==1)) {
		  /* Save for folding */
		  // need to add neg DM folding also.
		  memcpy(&client.fold_buf[state.fold_buf_loc], power,
			 sizeof(float)*(state.fft_len>>(l+1)));
		  state.fold_buf_loc += (state.fft_len>>(l+1));
		}
		coadd_in_place(power, state.fft_len>>(l+1));
	      } // coadd level (l)
	      client.checkpoint();
	    } // dm sign (s)
	    state.dm_sign = 1;
	  } // inner dm (k)
	  state.dm_now = 0;
#ifdef BOINC_APP_GRAPHICS
	  if (abp) {
	    ap_graphics.generate_done(abp);
	  }
#endif
	} // data (j)

	if (state.fold_buf_loc != state.fold_buf_size) {
	  /*fprintf(stderr, "Fold buffer didn't fill up (lol=%d, size=%d).\n",
	    state.fold_buf_loc, state.fold_buf_size);*/
	}
	state.fold_buf_loc = 0;
	state.data_chunk_now = 0;

	// do fold - results output within ffa
	ffa(&client.fold_buf[0], (float)1e-3);

      } // DM chunk (i)

        // output the top result no matter how many we have >= threshold, so
        // that if none >= threshold we can still validate redundant results
        // against each other.

    client.outfile.finish_results();
  }

  void Client::finish()
  {
    // Clean up stuff and exit
    // unloadGl();
    science.finish();
    outfile.finish();
    statefile.finish();
    foldfile.finish();
  }

}
