// Copyright 2003 Regents of the University of California

// Astropulse is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2, or (at your option) any later
// version.

// Astropulse is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.

// You should have received a copy of the GNU General Public License along
// with Astropulse; see the file COPYING.  If not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

/* ap_fileio.C */
/* Functions for reading data, etc for AstroPulse client */
#include "config.h"

#ifdef _WIN32
#include "windows.h"
#endif

#include <cstdio>
#include <cstdlib>
#include <vector>
#include <string>
#include <iostream>
#include <sstream>
#include <algorithm>

#include "parse.h"
#include "xml_util.h"
#include "db_table.h"
#include "db/ap_schema.h"
#include "astropulse.h"
#include "util.h"


#include <cstring>
#include <cassert>
#include <cstdlib>
#include <cstdio>
#include <cerrno>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif


namespace Astropulse
{
    const int AP_OUTPUT_PROTOCOL_VERSION = 3;

    // Filenames
    const char* WU_FNAME    = "in.dat";
    const char* OUT_FNAME   = "pulse.out";
    const char* STATE_FNAME = "ap_state.dat";
    const char* FOLD_FNAME  = "fold.dat";

    inline size_t get_file_size(const char* filename)
    {
        FILE* f = fopen(filename, "r");
        if (!f) return 0;
        fseek(f, 0, SEEK_END);
        size_t size = ftell(f);
        fclose(f);
        return size;
    }

    void Outfile::init()
    {
        client.status = "Opening output file.";

        string resolved_filename;
        boinc_resolve_filename_s(OUT_FNAME, resolved_filename);

        size_t size = get_file_size(resolved_filename.c_str());

        if (outfile.open(resolved_filename.c_str(), "a")) {
            fprintf(stderr, "Error opening outfile %s: %s\n", resolved_filename.c_str(), strerror(errno));
            exit(FOPEN_ERR);
        }

        // consistency checking
        if (client.statefile.resumed) {
            if (size == 0) {
                fprintf(stderr,
                        "Error resuming: statefile exists, but output file doesn't exist.\n");
                exit(FREAD_ERR);
            }
        } else {
            if (size != 0) {
                fprintf(stderr,
                        "Error initializing: statefile doesn't exist, but output file exists.\n");
                exit(FREAD_ERR);
            }

            outfile.flush();
        }
    }

    // Read in state file to resume computation
    void Statefile::read()
    {
        client.status = "Reading state file.";

        boinc_time_to_checkpoint();

        FILE* statefile = fopen(STATE_FNAME, "rb");
        if (statefile == NULL) return;

        int nread = fread(&client.state, 1, sizeof(state_t), statefile);
        if (nread != sizeof(state_t)) {
            fprintf(stderr,
                    "Error reading from statefile: wanted %d bytes, got %d\n",
                    sizeof(state_t), nread
                );
            exit(FREAD_ERR);
        }
        fclose(statefile);

        resumed = true;
        if (boinc_is_standalone()) {
            fprintf(stderr, "## Warning: resuming in standalone mode\n");
        }

        return;
    }

    // Write state file so we can resume later
    void Statefile::write()
    {
        client.status = "Saving computation state.";

        boinc_time_to_checkpoint();

        FILE* statefile = fopen(STATE_FNAME, "wb");
        if (statefile == NULL) {
            fprintf(stderr, "Error opening statefile for write: %s.\n", strerror(errno));
            exit(FOPEN_ERR);
        }

        int nwrite = fwrite(&client.state, sizeof(state_t), 1, statefile);
        if (nwrite != 1) {
            fprintf(stderr, "Error writing to statefile: %s.\n", strerror(errno));
            exit(FWRITE_ERR);
        }

        fclose(statefile);
    }

    void Foldfile::write()
    {
        client.status = "Saving fold buffer.";

        boinc_time_to_checkpoint();

        FILE* foldfile = fopen(FOLD_FNAME, "wb");
        if (foldfile == NULL) {
            fprintf(stderr, "Error opening foldfile for write: %s.\n", strerror(errno));
            exit(FOPEN_ERR);
        }

        int nwrite = fwrite(&(client.fold_buf[0]), sizeof(float), client.state.fold_buf_size, foldfile);
        if (nwrite != client.state.fold_buf_size) {
            fprintf(stderr, "Error writing to foldfile: %s.\n", strerror(errno));
            exit(FWRITE_ERR);
        }

        fclose(foldfile);
    }

    // Read saved folding array
    int Foldfile::read()
    {
        client.status = "Reading fold buffer.";
        boinc_time_to_checkpoint();

        FILE* foldfile = fopen(FOLD_FNAME, "rb");
        if (foldfile == NULL) { return(1); }

        int nread = fread(&(client.fold_buf[0]), sizeof(float), client.state.fold_buf_size, foldfile);
        if (nread != client.state.fold_buf_size) {
            fprintf(stderr, "Error reading from foldfile: wanted %d bytes, but read only %d bytes.\n", sizeof(float)*client.state.fold_buf_size, nread);
            return(1);
        }

        fclose(foldfile);
        return(0);
    }

    void Wufile::init()
    {
        client.status = "Opening WU file.";
        string resolved_filename;
        boinc_resolve_filename_s(WU_FNAME, resolved_filename);

        wufile.open(resolved_filename.c_str(),std::ios_base::binary);
        if (!wufile) {
            fprintf(stderr, "Error opening wufile %s: %s\n", resolved_filename.c_str(), strerror(errno));
            exit(FOPEN_ERR);
        }
    }

    // Parse header from wufile
    void Wufile::parse_header()
    {
        workunit_header &wuheader(client.wuheader);
	
        int parse_ok=0;

        state_t& state = client.state;

    wufile.seekg(0,std::ios::beg);
	wufile >> wuheader;
	if (wufile) {
	  parse_ok++;
	  state.datasize=wuheader.data_desc.nsamples*wuheader.recorder_cfg->bits_per_sample/8;
	  state.max_coadd=wuheader.analysis_cfg->max_coadd;
	  state.dm_low=wuheader.analysis_cfg->dm_low;
	  state.dm_hi=wuheader.analysis_cfg->dm_high;
	  state.dm_chunk=wuheader.analysis_cfg->dm_chunk;
	  state.fft_len=wuheader.analysis_cfg->fft_len;
	  state.ra0=wuheader.data_desc.start.ra;
	  state.dec0=wuheader.data_desc.start.dec;
	  state.jd0=wuheader.data_desc.start.time;
	  state.ra1=wuheader.data_desc.end.ra;
	  state.dec1=wuheader.data_desc.end.dec;
	  state.jd1=wuheader.data_desc.end.time;
	  for (unsigned int i=0; i<wuheader.analysis_cfg->thresholds.size(); i++) {
	    state.thresh[wuheader.analysis_cfg->thresholds[i].coadds]=
	       wuheader.analysis_cfg->thresholds[i].threshold;
	  }
        }

        if (parse_ok == 0) {
            fprintf(stderr, "Error parsing WU header.\n");
            exit(PARSE_ERR);
        }

        // Check for inconsistent/bad values
        if (state.fft_len > state.datasize*4) {
            fprintf(stderr, "Error parsing WU header: fft_len=%d > datasize=%d * 4\n",
                    state.fft_len, state.datasize);
            exit(PARSE_ERR);
        }
        if (state.dm_hi <= state.dm_low) {
            fprintf(stderr, "Error parsing WU header: dm_hi=%d <= dm_low=%d\n",
                    state.dm_hi, state.dm_low);
            exit(PARSE_ERR);
        }
        if (state.max_coadd > MAX_COADD) {
            fprintf(stderr, "Error parsing WU header: max_coadd=%d > %d\n",
                    state.max_coadd, MAX_COADD);
            exit(PARSE_ERR);
        }
        if ((state.dm_chunk>>state.max_coadd) > 1) {
            fprintf(stderr, "Error parsing WU header: (dm_chunk=%d) >> (max_coadd=%d) > 1\n",
                    state.dm_chunk, state.max_coadd);
            exit(PARSE_ERR);
        }
        if (state.fft_len < 0) {
            fprintf(stderr, "Error parsing WU header: fft_len=%d < 0\n",
                    state.fft_len);
            exit(PARSE_ERR);
        }
        if (state.datasize < 0) {
            fprintf(stderr, "Error parsing WU header: datasize=%d < 0\n",
                    state.datasize);
            exit(PARSE_ERR);
        }
        if (state.max_coadd < 0) {
            fprintf(stderr, "Error parsing WU header: max_coadd=%d < 0\n",
                    state.max_coadd);
            exit(PARSE_ERR);
        }
        if (state.dm_hi < 0) {
            fprintf(stderr, "Error parsing WU header: dm_hi=%d < 0\n",
                    state.dm_hi);
            exit(PARSE_ERR);
        }
        if (state.dm_low < 0) {
            fprintf(stderr, "Error parsing WU header: dm_low=%d < 0\n",
                    state.dm_low);
            exit(PARSE_ERR);
        }

        client.outfile.write_header();
    }

    // returns true on success
    inline bool find_start_tag(FILE* file, const char* tag)
    {
        rewind(file);
        char buf[256];
        while (fgets(buf, 256, file)) {
            if (match_tag(buf, tag)) return true;
        }
        return false;
    }

    void Wufile::read_raw_data()
    {
        client.status = "Reading raw data.";
        boinc_time_to_checkpoint();
	std::string data_chunk;
	data_chunk.reserve(client.state.datasize+1024);
	wufile.unsetf(std::ios_base::skipws);

        while (!wufile.eof()) {
	  char s;
	  wufile >> s;
	  data_chunk+=s;
	}

	client.rawdata=xml_decode_field<unsigned char>(data_chunk,(const char *)"data");
	
        if (client.rawdata.size() != client.state.datasize) {
            fprintf(stderr, "Error reading raw data: Expected %d bytes, read %d.\n", client.state.datasize, client.rawdata.size());
            exit(FREAD_ERR);
        }
    }

    void Wufile::finish()
    {
        wufile.close();
    }

    // encode the time series as a ASCII integers

    void Outfile::write_header()
    {
        outfile.printf("<astropulse_output>\n");
        std::string s(client.wuheader.print_xml());
        outfile.printf("%s",s.c_str());
	outfile.printf("<pulses>\n");
    }

    void Outfile::output_result(const ap_signal &sig) 
    {
        std::string s(sig.print_xml(0,0,1));
        outfile.printf("%s",s.c_str());
        // REMINDER: do not flush here; only do so at checkpoint.  All state
        // files have to be consistent in case of crash.
    }

    // write best results for each coadd level (we do this for the sake of
    // the validator).
    void Outfile::finish_results()
    {
        outfile.printf("</pulses>\n");
        outfile.printf("<best_pulses>\n");
        ap_signal *best=client.state.best;
        for (int l = 0; l < client.state.max_coadd; ++l) {
            assert(best[l].peak_power > -INFINITY);
            std::string s(best[l].print_xml(0,0,1));
            outfile.printf("%s",s.c_str());
        }
        outfile.printf("</best_pulses>\n");
        outfile.printf("</astropulse_output>\n");
    }

    void Outfile::finish()
    {
        outfile.flush();
        outfile.close();
    }

    void Statefile::finish()
    {
        unlink(STATE_FNAME);
    }

    void Foldfile::finish()
    {
        unlink(FOLD_FNAME);
    }
}
