// Copyright 2003 Regents of the University of California

// Astropulse is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2, or (at your option) any later
// version.

// Astropulse is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.

// You should have received a copy of the GNU General Public License along
// with Astropulse; see the file COPYING.  If not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

/* ap_science.C */
/* Functions for AstroPulse scientific computations */

#ifndef _WIN32
#include "config.h"
#endif

#include <cmath>
#include "astropulse.h"
#include "boinc_api.h"
#include "sbtf.h"


/* Windows is stupid */
#ifndef M_PI
#define M_PI 3.141592653589793238462
#endif

namespace Astropulse {
    const char* WISDOM_FNAME =  "wisdom.dat";

    inline void Science::Wisdom::load() {
        // Load previous wisdom
        FILE *wisdomfile= fopen(WISDOM_FNAME, "r");
        if (wisdomfile) {
            fftwf_import_wisdom_from_file(wisdomfile);
            fclose(wisdomfile);
        }
    }

    inline void Science::Wisdom::save() {
        // Save new wisdom
        FILE* wisdomfile = fopen(WISDOM_FNAME, "w");
        if (wisdomfile) {
            fftwf_export_wisdom_to_file(wisdomfile);
            fclose(wisdomfile);
        }
    }

    // Allocate memory, and create fft plans
    void Science::init()
    {
        const size_t& fft_len = client.state.fft_len;

        data  = (fftwf_complex *)fftwf_malloc(sizeof(fftwf_complex)*fft_len);
        temp  = (fftwf_complex *)fftwf_malloc(sizeof(fftwf_complex)*fft_len);
        chirp  = (fftwf_complex *)fftwf_malloc(sizeof(fftwf_complex)*fft_len*client.state.dm_chunk);

        client.status = "Creating FFT plans.";
        boinc_time_to_checkpoint();

        wisdom.load();

        client.status = "Creating forward FFT plan.";

        fp = fftwf_plan_dft_1d(fft_len, data, data, FFTW_FORWARD, FFTW_MEASURE);

        client.status = "Creating inverse FFT plan.";
        ip = fftwf_plan_dft_1d(fft_len, temp, temp, FFTW_BACKWARD, FFTW_MEASURE);

        wisdom.save();

        boinc_time_to_checkpoint();
    }

    void Science::finish()
    {
        fftwf_destroy_plan(fp);
        fftwf_destroy_plan(ip);

        fftwf_free(chirp);
        fftwf_free(temp);
        fftwf_free(data);
    }

    // Build dechirp table
    void Science::build_chirp_table(int dm_start) {
        int i,j;
        double freq;
        double phase;

        client.status = "Building chirp table.";
        boinc_time_to_checkpoint();

        state_t& state = client.state;

        for (i=0; i<client.state.dm_chunk; i++) {
            for (j=0; j<client.state.fft_len; j++) {
                freq = (double)j/(double)client.state.fft_len;
                freq = (j<state.fft_len/2)?freq:freq-1.0;
                phase = 2.0*M_PI*(double)(dm_start+i)*freq*(1.0+freq);
                chirp[i*state.fft_len + j][0] = (float)cos(phase);
                chirp[i*state.fft_len + j][1] = (float)sin(phase);
            }
        }

        boinc_time_to_checkpoint();
    }

    void Science::convert_bits_to_float() {
        client.status = "Converting raw data.";
        boinc_time_to_checkpoint();
        if (splitter_bits_to_float(&client.rawdata[client.state.data_chunk_now],
                                   (float *)data, client.state.fft_len)) {
            exit(SBTF_ERR);
        }
    }

    void Science::compute_forward_fft() {
        client.status = "Computing forward FFT.";
        boinc_time_to_checkpoint();
        fftwf_execute(fp);
    }

    void Science::dechirp(int dm, float *power, int sign) {
        const size_t& fft_len = client.state.fft_len;
        int n = dm*fft_len;
        size_t i;
        static float avg_pow = 2*fft_len*float(fft_len);
        // avg_pow contains N^2 for unnormalized fft, and 2 for 1-bit
        // convention

        client.status = "Dechirping.";
        boinc_time_to_checkpoint();

        for (i=0; i<fft_len; i++) {
            /* positive DM */
            if (sign >= 0) {
                temp[i][0] = data[i][0]*chirp[n+i][0] - data[i][1]*chirp[n+i][1];
                temp[i][1] = data[i][1]*chirp[n+i][0] + data[i][0]*chirp[n+i][1];
            }
            /* negative DM */
            else {
                temp[i][0] = data[i][0]*chirp[n+i][0] + data[i][1]*chirp[n+i][1];
                temp[i][1] = data[i][1]*chirp[n+i][0] - data[i][0]*chirp[n+i][1];
            }
        }

        client.status = "Computing inverse FFT.";
        boinc_time_to_checkpoint();
        fftwf_execute(ip);

        /* only save 1/2 of the fft ... need to check this */
        for (i=0; i<fft_len/2; i++) {
            client.power[i] = (float)((temp[i][0]*temp[i][0] + temp[i][1]*temp[i][1])/avg_pow);
        }
    }

    void Science::coadd_in_place(float *power, int size) {
        int i;

        for (i=0; i<size/2; i++) {
            power[i] = power[2*i] + power[2*i+1];
        }
    }
}
