// Copyright 2003 Regents of the University of California

// Astropulse is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2, or (at your option) any later
// version.

// Astropulse is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.

// You should have received a copy of the GNU General Public License along
// with Astropulse; see the file COPYING.  If not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifndef H_ASTROPULSE
#define H_ASTROPULSE

#include "ctype.h"
#include "stdio.h"
#include <string>
#include "fftw3.h"
#include "boinc_api.h"
#include "mfile.h"
#include "db/db_table.h"
#include "db/ap_schema.h"

namespace Astropulse
{
    using std::string;

    static const int MAX_COADD = 15;

    // Structure which includes everything that might be in a wu header.  Also
    // contains info for state file.
    struct state_t {
      int datasize;           // Size of raw data in bytes
      int dm_low;             // Lowest dm to calculate
      int dm_hi;              // Higest dm
      int dm_chunk;           // dm resolution for folding - also dm table size
      int fft_len;            // fft length >= 2*dm_hi
      int max_coadd;          // max level to collapse down to ( >= log(dm_chunk) )
      int fold_buf_size;      // size of array to fold
      int fold_level;         // should be log2(dm_chunk)
      double thresh[MAX_COADD];// array of thresholds (max_collapse total)
      double ra0;             // RA at start of wu, in hours
      double ra1;             // RA at end of wu
      double dec0;            // Dec at start of wu, in degrees
      double dec1;            // Dec at end of wu
      double jd0;             // JD at start
      double jd1;             // JD at end
      int data_chunk_now;     // Current location being worked on
      int dm_chunk_now;       // Current dm chunk
      int dm_now;             // Current dm within chunk
      int dm_sign;            // positive (+1) or negative (-1) dispersion
      int fold_buf_loc;       // location to write in fold buffer
      double frac_done;       // fraction of wu completed so far
      double pulse_jd;        // dummy to calculate the julian date of pulse
      int result_count;       // number of results found so far
      ap_signal best[MAX_COADD];
      
      state_t();
      void compute_fraction_done(double l);
    };

    struct time_series_f_t {
        enum { MAXLENGTH = 256 };
	std::vector<float> data;
        int length;

        time_series_f_t() : length(0) { data.reserve(MAXLENGTH); }
        void build(float array[], int index, int range);
    };

    struct time_series_b_t {
        enum { MAXLENGTH = 64 };
	std::vector<unsigned char> data;
        int length;

        time_series_b_t() : length(0) { data.reserve(MAXLENGTH); }
        void build(float array[], int index, int range);
    };

    enum ERROR_CODES {
        FWRITE_ERR = -1,                      // Error writing a file.
        SBTF_ERR  = -2,                       // Error from splitter_bits_to_float
        FOPEN_ERR = -3,
        PARSE_ERR = -4,
        FREAD_ERR = -5
    };

    class Wufile
    {
        std::ifstream wufile;
    public:
        void init();
        void parse_header();
        void read_raw_data();
        void finish();
    };

    class Outfile
    {
      MFILE outfile;
    public:
        void init();
        void finish_results();
        void finish();

        void output_result(const ap_signal &s);
        void write_header();
        void flush() { outfile.flush(); }
    };

    class Statefile
    {
    public:
        bool resumed;
        Statefile() : resumed(0) {}
        void read();
        void write();
        void finish();
    };

    class Foldfile
    {
    public:
        int read();
        void write();
        void finish();
    };

    class Science
    {
        fftwf_complex* data;
        fftwf_complex* temp;
        fftwf_complex* chirp;
        fftwf_plan fp, ip;

        class Wisdom {
        public:
            void load();
            void save();
        };
        Wisdom wisdom;
    public:
        void init();
        void mainloop();
        void finish();
    protected:
        void build_chirp_table(int dm_start);
        void convert_bits_to_float();
        void compute_forward_fft();
        void dechirp(int dm, float *power, int sign);
        void coadd_in_place(float *power, int size);
        int ffa(float *data, float dp);
    };

    class Client {
    public:
        workunit_header wuheader;
        state_t state;
        const char* status;         // Spot to put program status for display.
        // NOTE: any data assigned to `status' must be static data -- if this
        // constraint is not acceptable, change it back to `string' instead
        // of `const char*'.

        Outfile   outfile;
        Wufile    wufile;
        Foldfile  foldfile;
        Statefile statefile;

        Science science;

	std::vector<unsigned char> rawdata;
	std::vector<float> power;
	std::vector<float> fold_buf;

        void init();
        void finish();

        inline void do_checkpoint();
        inline void checkpoint();
    };

    extern Client client;
};

#ifndef INFINITY
#define INFINITY (-log(0.0))
#endif

#endif

